/*
	DbFS.NET
    Copyright (C) 2016 SCAND Ltd, svg@scand.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

namespace dbfs
{
	using System;
	using System.Collections.Generic;
	using System.IO;
	using dbfs.model;
	using dbfs.repository;
	using dbfs.utils;

	public class CreateCommit : IDisposable
	{
		internal CreateCommit(DBFS dbfs, int lock_timeout, int branch_id)
		{
			_dbfs = dbfs;
			_lock_timeout = lock_timeout;
			_db = _dbfs.DB();

			_branch = _db.Branch.Get(branch_id);

			_state = new State(_db, branch_id);
		}

		# region message and user

		private string _message;
		private model.User _user;

		public string Message
		{
			get
			{
				if (_disposed)
					throw new ObjectDisposedException("this");

				return _message;
			}
			set
			{
				if (_disposed)
					throw new ObjectDisposedException("this");

				_message = value;
			}
		}

		public CreateCommit SetMessage(string message)
		{
			Message = message;

			return this;
		}

		public CreateCommit SetUser(string name)
		{
			if (_disposed)
				throw new ObjectDisposedException("this");

			_user = _db.User.Get(name);

			return this;
		}

		#endregion

		#region source

		public IFileSource Source(File file)
		{
			return new FileSource { File = file };
		}

		public IDirectorySource Source(Folder folder)
		{
			var branch = _db.Branch.GetByCommitId(folder.CommitId);

			return new DirectorySource { Folder = folder, BranchID = branch.id, CommitID = folder.CommitId };
		}

		/// <summary>
		/// Use file/folder by path from current branch as source.
		/// </summary>
		public ISource Source(string path)
		{
			return new Source { Path = path, BranchID = _branch.id };
		}

		public ISource Source(string path, string branch_name)
		{
			var branch = _db.Branch.Get(branch_name);

			return new Source { Path = path, BranchID = branch.id };
		}

		public ISource Source(string path, int commit_id)
		{
			var branch = _db.Branch.GetByCommitId(commit_id);

			return new Source { Path = path, BranchID = branch.id, CommitID = commit_id };
		}

		#endregion

		#region public API for files

		public void AddFile(string path, Func<string, Stream> func)
		{
			if (_disposed)
				throw new ObjectDisposedException("this");
			// todo: check path

			_state.AddFile(path, new AsyncStream(func));
		}

		public void CopyFile(ISource source, string path)
		{
			if (_disposed)
				throw new ObjectDisposedException("this");

			_state.CopyFile((Source) source, path);
		}

		public void CopyFileTo(ISource source, string folder)
		{
			if (_disposed)
				throw new ObjectDisposedException("this");

			var source_ = (Source) source;

			_state.CopyFile(source_, FileUtils.CombinePath(folder, source_.Name()));
		}

		public void RenameFile(string source, string target)
		{
			if (_disposed)
				throw new ObjectDisposedException("this");

			_state.RenameFile(source, target);
		}

		public void MoveFileTo(string file, string folder)
		{
			if (_disposed)
				throw new ObjectDisposedException("this");

			var new_ = FileUtils.CombinePath(folder, FileUtils.GetFileName(file));

			RenameFile(file, new_);
		}

		public void RemoveFile(string path)
		{
			if (_disposed)
				throw new ObjectDisposedException("this");

			_state.RemoveFile(path);
		}

		#endregion

		#region public API for folders

		public void AddFolder(string path)
		{
			if (_disposed)
				throw new ObjectDisposedException("this");

			_state.AddFolder(path.EnsureFolder());
		}

		public void CopyFolder(ISource source, string path)
		{
			if (_disposed)
				throw new ObjectDisposedException("this");

			var source_ = (Source) source;
			if (source_.Path != null) source_.Path = source_.Path.EnsureFolder();

			_state.CopyFolder((Source) source, path.EnsureFolder());
		}

		public void CopyFolderTo(ISource source, string folder)
		{
			if (_disposed)
				throw new ObjectDisposedException("this");

			CopyFolder(source, FileUtils.CombinePath(folder, ((Source) source).Name()));
		}

		public void RenameFolder(string source, string target)
		{
			if (_disposed)
				throw new ObjectDisposedException("this");

			_state.RenameFolder(source.EnsureFolder(), target.EnsureFolder());
		}

		public void MoveFolderTo(string source, string folder)
		{
			if (_disposed)
				throw new ObjectDisposedException("this");

			var new_ = FileUtils.CombinePath(folder, FileUtils.GetFileName(source));

			RenameFolder(source, new_);
		}

		public void RemoveFolder(string path)
		{
			if (_disposed)
				throw new ObjectDisposedException("this");

			_state.RemoveFolder(path.EnsureFolder());
		}

		#endregion

		#region Cancel and Finish

		/// <summary>Automatically calls Dispose method.</summary>
		public void Cancel()
		{
			this.Dispose();
		}

		/// <summary>
		/// Process all operations and finish commit. Returns null if no changes was made.
		/// </summary>
		public Commit Finish()
		{
			if (_disposed)
				throw new ObjectDisposedException("this");
			if (_user == null)
				throw new InvalidOperationException("user is not specified");

			_db.Branch.Lock(_branch, _lock_timeout, _user.id);

			Commit wrapped_commit = null;
			bool is_failed = true;

			try
			{
				var commit = finish();

				if (commit != null)
					wrapped_commit = new Commit(commit, _branch);
					
				is_failed = false;
			}
			finally {
				if (!is_failed) {
					_db.Branch.Unlock(_branch);
					_dbfs._reload();
				}
				else {
					using (var db = _dbfs.DB())
					{
						db.Branch.Unlock(db.Branch.Get(_branch.id));
					}
				}
			}

			return wrapped_commit;
		}

		#endregion

		#region private methods

		private dbfs.model.Commit finish()
		{
			var commit = _db.Commit.Create(_message, _user);

			_state.Operations
				.ForEach(o => {
					if (o.astream != null) {
						var blob = upload(o.revision.File.path, o.astream, o.previous_blob._(e => e.sha1));
						if (blob == null) return;

						o.revision.Blob = blob;
					}

					if (o.latest == null)
						_db.Latest.Create(_branch, o.revision);
					else
						o.latest.FileRevision = o.revision;

					o.revision.Commit = commit;
					commit.FileRevisions.Add(o.revision);
				})
			;

			if (commit.FileRevisions.Count != 0) {
				_db.Commit.CreateRelation(_branch, commit);
				_branch.LatestCommit = commit;

				_db.SaveChanges();

				return commit;
			}

			return null;
		}

		private Blob upload(string path, AsyncStream astream, string previous_hash = null)
		{
			Stream stream = null;
			string tmpfile = null;

			try {
				stream = astream(path);
				string sha1 = calculateSHA1(ref stream, ref tmpfile);

				if (previous_hash != null && sha1 == previous_hash)
					return null;

				return _db.Blob.TryGet(sha1) ?? _db.Blob.Save(stream);
			}
			finally {
				if (stream != null) stream.Dispose();
				if (tmpfile != null) System.IO.File.Delete(tmpfile);
			}
		}

		private string calculateSHA1(ref Stream stream, ref string tmpfile)
		{
			if (stream.CanSeek) {
				long origin = stream.Position;

				var hash = CryptUtils.ComputeHash(stream);

				stream.Seek(origin, SeekOrigin.Begin);

				return hash;
			}
			else {
				tmpfile = Path.GetTempFileName();

				var tmp = tmpfile; // for lambda
				var hash = CryptUtils.ComputeHashWhile(stream, (stream_) =>
					{
						// todo: can be slow
						FileUtils.SaveToFile(stream_, tmp);
					}
				);

				stream.Dispose();

				stream = System.IO.File.OpenRead(tmpfile);

				return hash;
			}
		}

		#endregion

		#region IDisposable implementation

		private bool _disposed = false;

		public void Dispose()
		{
			if (!_disposed)
			{
				_branch = null;
				_user = null;
				_dbfs = null;

				if (_db != null) _db.Dispose();
				_db = null;

				_disposed = true;
			}
		}

		#endregion

		private DBFS _dbfs;
		private int _lock_timeout;
		private Repository _db;

		private model.Branch _branch;

		private State _state;
	}
}
