﻿/*
	DbFS.NET
    Copyright (C) 2016 SCAND Ltd, svg@scand.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

namespace dbfs.model
{
	using System;
	using System.Collections.Generic;
	using System.IO;
	using System.Linq;
	using repository;
	using utils;

	public delegate Stream AsyncStream(string path);

	internal class State
	{
		internal State(Repository db, int branch_id)
		{
			_db = db;
			_branch = branch_id;
		}

		public IEnumerable<Operation> Operations
		{
			get { return _state.Values.Where(e => e.revision != null); }
		}

		#region file

		public void AddFile(string path, AsyncStream astream)
		{
			AddFolder(FileUtils.GetDirectoryName(path).EnsureFolder());

			Operation o = operation(path);

			o.revision = new FileRevision {
				File = o.latest._(e => e.File) ?? _db.File.Create(path),
				Type = o.latest.Exist() ? ChangeType.UPDATE : ChangeType.ADD,
			};

			o.astream = astream;
			o.previous_blob = o.latest._(e => e.FileRevision.Blob);
		}

		public void CopyFile(Source source, string path)
		{
			AddFolder(FileUtils.GetDirectoryName(path).EnsureFolder());

			Operation o = operation(path);

			var info = source.GetInfo(_db);

			if (o.latest != null
				&& !o.latest.FileRevision.IsRemoved()
				&& info.Blob.id == o.latest.FileRevision.blob_id)
			{
				return;
			}

			o.revision = new FileRevision {
				File = o.latest._(e => e.File) ?? _db.File.Create(path),
				Type = ChangeType.COPY,
				Blob = info.Blob,
				source_id = info.RevisionID,
			};

			o.Clear();
		}

		public void RenameFile(string source, string target)
		{
			AddFolder(FileUtils.GetDirectoryName(target).EnsureFolder());

			Operation renamed = operation(source);

			if (renamed.revision != null)
				throw new NotImplementedException("simultaneous operations");

			if (!renamed.latest.Exist()) return;

			Operation rename = operation(target);

			if (rename.latest == null
				|| rename.latest.FileRevision.IsRemoved()
				|| rename.latest.FileRevision.blob_id != renamed.latest.FileRevision.blob_id)
			{
				rename.revision = new FileRevision {
					File = rename.latest._(e => e.File) ?? _db.File.Create(target),
					Type = ChangeType.RENAME,
					Blob = renamed.latest.FileRevision.Blob,
					source_id = renamed.latest.revision_id,
				};

				rename.Clear();
			}

			renamed.revision = new FileRevision {
				File = renamed.latest.File,
				Type = ChangeType.RENAMED,
			};

			renamed.Clear();
		}

		public void RemoveFile(string path)
		{
			Operation o = operation(path);

			if (!o.latest.Exist()) {
				o.revision = null;
				o.Clear();

				return;
			}

			o.revision = new FileRevision {
				File = o.latest.File,
				Type = ChangeType.DELETE,
			};

			o.Clear();
		}

		#endregion

		#region folder

		public void AddFolder(string path)
		{
			Helpers.GetAllFoldersFromSinglePath(path)
				.Reverse()
				.ForEach((folder, stop) => {
					Operation o = operation(folder);

					// no need to readd existing folder
					if (o.latest.Exist()) {
						if (o.revision == null)
							stop();
						else
							o.revision = null;

						return;
					}

					o.revision = new FileRevision {
						File = o.latest._(e => e.File) ?? _db.File.Create(folder),
						Type = o.latest.Exist() ? ChangeType.UPDATE : ChangeType.ADD,
					};
				})
			;
		}

		public void CopyFolder(Source source, string path)
		{
			var branch = _db.Branch.Get(source.BranchID);

			var to_update = _db.Latest.AllNested(_branch, path).ToDictionary(e => e.File.path);

			_db.Revision.AllNestedOptimized(branch, source.GetPath(), source.CommitID, true)
				.ForEach(copy => {
					string target_path = copy.File.path.Replace(source.GetPath(), path);

					LatestRevision latest = null;
					to_update.TryGetValue(target_path, out latest);

					Operation o = operation(latest, target_path);

					if (latest != null
						&& copy.blob_id == latest.FileRevision.blob_id
						&& copy.blob_id != null) return;

					o.revision = new FileRevision {
						File = o.latest._(e => e.File) ?? _db.File.Create(target_path),
						Type = ChangeType.COPY,
						Blob = copy.Blob,
						source_id = copy.id,
					};

					o.Clear();
				})
			;
		}

		public void RenameFolder(string source, string target)
		{
			var to_update = _db.Latest.AllNested(_branch, target).ToDictionary(e => e.File.path);

			_db.Latest.AllNested(_branch, source, true)
				.ForEach(l_renamed => {
					Operation renamed = operation(l_renamed);

					if (renamed.revision != null)
						throw new NotImplementedException("simultaneous operations");

					string target_path = l_renamed.File.path.Replace(source, target);

					LatestRevision l_rename = null;
					to_update.TryGetValue(target_path, out l_rename);

					Operation rename = operation(l_rename, target_path);

					if (l_rename == null
						|| l_rename.FileRevision.IsRemoved()
						|| l_rename.FileRevision.blob_id != l_renamed.FileRevision.blob_id)
					{
						rename.revision = new FileRevision {
							File = rename.latest._(e => e.File) ?? _db.File.Create(target_path),
							Type = ChangeType.RENAME,
							Blob = renamed.latest.FileRevision.Blob,
							source_id = renamed.latest.revision_id,
						};

						rename.Clear();
					}

					renamed.revision = new FileRevision {
						File = renamed.latest.File,
						Type = ChangeType.RENAMED,
					};

					renamed.Clear();
				})
			;
		}

		public void RemoveFolder(string path)
		{
			// remove added items from current state
			// todo: for memory optimization you can remove only those records which is new for storage
			_state.RemoveWhere(e => e.StartsWith(path));

			_db.Latest.AllNested(_branch, path, true)
				.ForEach(latest => {
					Operation o = operation(latest);

					o.revision = new FileRevision {
						File = o.latest.File,
						Type = ChangeType.DELETE,
					};

					o.Clear();
				})
			;
		}

		#endregion

		#region private methods

		private Operation operation(string path)
		{
			Operation operation;

			if (_state.TryGetValue(path, out operation)) return operation;

			return _state[path] = new Operation {
				latest = _db.Latest.TryGet(_branch, path)
			};
		}

		private Operation operation(LatestRevision latest, string path = null)
		{
			if (path == null) path = latest.File.path;

			Operation operation;

			if (_state.TryGetValue(path, out operation)) return operation;

			return _state[path] = new Operation {
				latest = latest
			};
		}

		#endregion

		Dictionary<string, Operation> _state = new Dictionary<string, Operation>();
		Repository _db;
		int _branch;


		public class Operation
		{
			public FileRevision revision;
			public LatestRevision latest;
			public AsyncStream astream;
			public Blob previous_blob;

			public void Clear()
			{
				this.astream = null;
				this.previous_blob = null;
			}
		}
	}

	static class StateExtensions
	{
		public static bool Exist(this LatestRevision latest)
		{
			return latest != null
				&& latest.FileRevision.Type != ChangeType.DELETE
				&& latest.FileRevision.Type != ChangeType.RENAMED
			;
		}
	}
}
