/*
	DbFS.NET
    Copyright (C) 2016 SCAND Ltd, svg@scand.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

namespace dbfs.dbfscmd
{
	using System;
	using System.Collections.Generic;
	using System.IO;
	using System.Linq;
	using ManyConsole;
	using dbfs.utils;

	public class CheckoutCommand : ConsoleCommand
	{
		public CheckoutCommand()
		{
			this.SkipsCommandSummaryBeforeRunning();
			this.IsCommand("checkout", "get files content");

			this.HasOption("c|commit=", "provide commit id", (int v) => commit_id = v);
			this.HasOption("b|branch=", "branch name", v => branch_name = v);

			this.HasAdditionalArguments(2, "<virtual-path> <real-path>");
		}

		private int commit_id = -1;
		private string branch_name;

		public override int Run(string[] remainingArguments)
		{
			this._checkArguments(remainingArguments);

			var dbfs = Program.GetDBFS();

			string virtual_path = remainingArguments[0];
			string real_path = remainingArguments[1];

			// fixme: hmm?
			bool folder = virtual_path[virtual_path.Length - 1] == '/';
			if (folder) {
				_saveFolderRecursively(dbfs, real_path, virtual_path);
			}
			else {
				var file = dbfs.GetFile(commit_id, virtual_path);

				string filename = real_path;
				if (real_path[real_path.Length - 1] == '/') {
					filename = FileUtils.CombinePath(real_path, file.Name);
				}

				dbfs.GetContent(file).SaveToFile(filename);
			}

			return 0;
		}

		private void _checkArguments(string[] remainingArguments)
		{
			bool commit = commit_id != -1;
			bool branch = !String.IsNullOrEmpty(branch_name);

			if (commit == branch)
				throw new ConsoleHelpAsException("only single command (--commit and --branch) allowed at the same time");
		}

		private void _saveFolderRecursively(dbfs.DBFS dbfs, string real_path, string virtual_path)
		{
			FileUtils.CreateDir(real_path);

			_files(dbfs, virtual_path)
				.ToList()
				.ForEach(file => dbfs.GetContent(file).SaveToFile(FileUtils.CombinePath(real_path, file.Name)))
			;

			_folders(dbfs, virtual_path)
				.ToList()
				.ForEach(folder =>
					_saveFolderRecursively(
						dbfs,
						FileUtils.CombinePath(real_path, folder.Name),
						FileUtils.CombinePath(virtual_path, folder.Name))
				)
			;
		}

		private IEnumerable<dbfs.File> _files(dbfs.DBFS dbfs, string directory)
		{
			if (commit_id != -1)
				return dbfs.GetFilesList(commit_id, directory);
			else
				return dbfs.GetFilesList(branch_name, directory);
		}

		private IEnumerable<dbfs.Folder> _folders(dbfs.DBFS dbfs, string directory)
		{
			if (commit_id != -1)
				return dbfs.GetFoldersList(commit_id, directory);
			else
				return dbfs.GetFoldersList(branch_name, directory);
		}
	}
}
