﻿/*
	DbFS.NET
    Copyright (C) 2016 SCAND Ltd, svg@scand.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

namespace dbfs.dbfscmd
{
	using System;
	using System.Collections.Generic;
	using System.IO;
	using ManyConsole;
	using dbfs.utils;

	public class CommitCommand : ConsoleCommand
	{
		public CommitCommand()
		{
			this.SkipsCommandSummaryBeforeRunning();
			this.IsCommand("commit", "create commit");

			this.HasRequiredOption<string>("m|message=", "provide commit message", v => message = v);
			this.HasRequiredOption<string>("user=", "provide user name", v => this._user = v);
			this.HasRequiredOption<string>("branch=", "provide branch name", v => this._branch = v);
			this.HasOption<string>("remove=", "provide virtual path to remove", v => m_args.Add(new Argument(Argument.Type.Remove, v)));
			this.HasOption<string>("add=", "provide path to add", v => m_args.Add(new Argument(Argument.Type.Add, v)));
			this.HasOption<string>("path=", "provide virtual path or path prefix", v => m_args.Add(new Argument(Argument.Type.Path, v)));
			this.HasOption("no-recursion", "", v => _recursion = v == null);

			this.HasOption<string>("copy=", "<from>[:<commit_id|branch_name>]::<to>", _copy);
			this.HasOption<string>("move=", "<from>::<to>", _move);

			this.HasOption("time", "", v => _time = v != null);
		}

		void _copy(string command)
		{
			Argument copy = new Argument { type = Argument.Type.Copy };

			var paths = command.Split(new string[] { "::" }, StringSplitOptions.None);
			if (paths.Length != 2) throw new ArgumentException();

			var sources = paths[0].Split(':');

			if (sources.Length == 2) {
				if (!int.TryParse(sources[1], out copy.commit_id))
					copy.branch_name = sources[1];
			}

			// copying folders...
			if (sources[0].EndsWith("/"))
				throw new NotImplementedException();

			copy.source = sources[0];

			if (paths[1].EndsWith("/"))
				copy.dest = FileUtils.CombinePath(paths[1], FileUtils.GetFileName(sources[0]));
			else
				copy.dest = paths[1];

			m_args.Add(copy);
		}

		void _move(string command)
		{
			Argument move = new Argument { type = Argument.Type.Move };

			var paths = command.Split(new string[] { "::" }, StringSplitOptions.None);
			if (paths.Length != 2) throw new ArgumentException();

			move.source = paths[0];
			move.dest = paths[1];

			m_args.Add(move);
		}

		private string message;
		private string _user;
		private string _branch;

		private bool _recursion = true;

		private bool _time = false;

		private List<Argument> m_args = new List<Argument>();
		private dbfs.CreateCommit m_commit;

		public override int Run(string[] remainingArguments)
		{
			this._checkArguments(remainingArguments);

			System.Diagnostics.Stopwatch watch = null;

			if (_time) watch = System.Diagnostics.Stopwatch.StartNew();

			var dbfs = Program.GetDBFS();
			dbfs.Commit result;

			{
				m_commit = dbfs.StartCommit(message, _user, _branch);

				Argument last = null;

				foreach (var arg in m_args) {
					switch (arg.type) {
						case Argument.Type.Remove:
							m_remove(arg.arg);
							break;
						case Argument.Type.Add:
							m_process(ref last, arg);
							break;
						case Argument.Type.Path:
							m_process(ref last, arg);
							break;
						case Argument.Type.Copy:
							dbfs.ISource source;

							if (arg.branch_name != null)
								source = m_commit.Source(arg.source, arg.branch_name);
							else if (arg.commit_id != 0)
								source = m_commit.Source(arg.source, arg.commit_id);
							else
								source = m_commit.Source(arg.source);

							m_commit.CopyFile(source, arg.dest);
							break;
						case Argument.Type.Move:
							if (arg.source.EndsWith("/"))
								m_commit.RenameFolder(arg.source, arg.dest);
							else if (arg.dest.EndsWith("/"))
								m_commit.MoveFileTo(arg.source, arg.dest);
							else
								m_commit.RenameFile(arg.source, arg.dest);
							break;
						default: throw new NotImplementedException();
					}
				}

				if (last != null)
					throw new ArgumentException();

				result = m_commit.Finish();
			}

			if (result != null) {
				Console.WriteLine("[{0} {1}] {2}", this._branch, result.Id, this.message);
				Console.WriteLine(" {0} file{1} changed", result.Files.Count, result.Files.Count != 1 ? "s" : "");
			}
			else {
				Console.WriteLine("nothing to commit, abort");
			}

			if (m_commit != null)
				m_commit.Dispose();

			if (_time) {
				watch.Stop();
				Console.WriteLine("Execution took {0} milliseconds", watch.ElapsedMilliseconds);
			}

			return 0;
		}

		private void _checkArguments(string[] remainingArguments)
		{
		}

		private void m_remove(string path)
		{
			if (path.EndsWith("/"))
				m_commit.RemoveFolder(path);
			else
				m_commit.RemoveFile(path);
		}

		private void m_process(ref Argument last, Argument current)
		{
			if (last == null) {
				last = current;
				return;
			}

			if (last.type == current.type)
				throw new ArgumentException();

			var real_path = last.type == Argument.Type.Add ? last.arg : current.arg;
			var virtual_path = last.type == Argument.Type.Path ? last.arg : current.arg;

			m_add(real_path, virtual_path);
			last = null;
		}

		private void m_add(string real_path, string virtual_path)
		{
			if (File.Exists(real_path)) {
				m_commit.AddFile(virtual_path, path => new FileStream(real_path, FileMode.Open));
			}
			else if (Directory.Exists(real_path)) {
				m_commit.AddFilesFromDirectory(real_path, virtual_path, this._recursion);
			}
			else {
				throw new FileNotFoundException("file doesn't exist", real_path);
			}
		}


		private class Argument
		{
			public Type type;
			public string arg;

			// for Copy and Move

			public string source;
			public string dest;

			// for Copy operation

			public string branch_name;
			public int commit_id;

			public Argument() {}

			public Argument(Type type, string arg)
			{
				this.type = type;
				this.arg = arg;
			}

			public enum Type
			{
				Remove,
				Add,
				Path,
				Copy,
				Move,
			}
		}
	}
}
