CREATE TABLE [dbo].[Users] (
	[id] INT NOT NULL IDENTITY(1,1) PRIMARY KEY,
	[UserName] NVARCHAR(250) NOT NULL UNIQUE,
	[UserPassword] NVARCHAR(250) NOT NULL
);
GO

IF NOT EXISTS(SELECT 1 FROM information_schema.schemata WHERE schema_name='dbfs')
	EXEC ('create schema [dbfs]');
GO

-- Dbfs tables

CREATE TABLE [dbfs].[Files] (
	[id] INT NOT NULL IDENTITY(1,1) PRIMARY KEY,

	[path] NVARCHAR(450) COLLATE Latin1_General_CS_AS NOT NULL UNIQUE, -- 450 is maximum for unique constraint
	[is_directory] BIT NOT NULL DEFAULT(0),
	[folder] NVARCHAR(450) NOT NULL -- using for fast search for files in one folder
);
GO

CREATE TABLE [dbfs].[Blobs] (
	[id] INT NOT NULL IDENTITY(1,1) PRIMARY KEY,

	[size] BIGINT NULL,
	[sha1] NVARCHAR(40) NULL,
	[content] VARBINARY(MAX) DEFAULT NULL
);

CREATE NONCLUSTERED INDEX [IX_Blobs__sha1]
on [dbfs].[Blobs]([sha1] ASC);

CREATE TABLE [dbfs].[Commits] (
	[id] INT NOT NULL IDENTITY(1,1) PRIMARY KEY,
	[user_id] INT NOT NULL,

	[message] NVARCHAR(MAX) NOT NULL,
	[date] DATETIME NOT NULL,

	CONSTRAINT [FK_Commits__user_id__Users] FOREIGN KEY ([user_id]) REFERENCES [dbo].[Users]([id])
);
GO

CREATE NONCLUSTERED INDEX [IX_Commits__user_id]
ON [dbfs].[Commits]([user_id] ASC);

CREATE TABLE [dbfs].[FileRevisions] (
	[id] INT NOT NULL IDENTITY(1,1) PRIMARY KEY,
	[file_id] INT NOT NULL,
	[blob_id] INT DEFAULT NULL,
	[commit_id] INT NOT NULL,

	[type] INT NOT NULL,
	[source_id] INT DEFAULT NULL, -- id of source revision (for COPY and RENAMED operations)

	CONSTRAINT [FK_FileRevisions__file_id__Files] FOREIGN KEY ([file_id]) REFERENCES [dbfs].[Files]([id]) ON DELETE CASCADE,
	CONSTRAINT [FK_FileRevisions__blob_id__Blobs] FOREIGN KEY ([blob_id]) REFERENCES [dbfs].[Blobs] ([id]) ON DELETE SET DEFAULT,
	CONSTRAINT [FK_FileRevisions__commit_id__Commits] FOREIGN KEY ([commit_id]) REFERENCES [dbfs].[Commits]([id]) ON DELETE CASCADE
);
GO

CREATE NONCLUSTERED INDEX [IX_FileRevisions__file_id]
ON [dbfs].[FileRevisions]([file_id] ASC);

CREATE NONCLUSTERED INDEX [IX_FileRevisions__blob_id]
ON [dbfs].[FileRevisions]([blob_id] ASC);

CREATE NONCLUSTERED INDEX [IX_FileRevisions__commit_id]
ON [dbfs].[FileRevisions]([commit_id] ASC);
GO

CREATE TABLE [dbfs].[Branches] (
	[id] INT NOT NULL IDENTITY(1,1) PRIMARY KEY,
	[user_id] INT NOT NULL,
	[origin_commit_id] INT NOT NULL,
	[latest_commit_id] INT NOT NULL,

	[name] NVARCHAR(256) NOT NULL UNIQUE,
	[date] DATETIME NOT NULL,

	[lock_user_id] INT DEFAULT NULL,
	[lock_release_date] DATETIME DEFAULT NULL,

	CONSTRAINT [FK_Branches__user_id__Users] FOREIGN KEY ([user_id]) REFERENCES [dbo].[Users]([id]),
	CONSTRAINT [FK_Branches__origin_commit_id__Commits] FOREIGN KEY ([origin_commit_id]) REFERENCES [dbfs].[Commits] ([id]),
	CONSTRAINT [FK_Branches__latest_commit_id__Commits] FOREIGN KEY ([latest_commit_id]) REFERENCES [dbfs].[Commits] ([id]),
	CONSTRAINT [FK_Branches__lock_user_id__Users] FOREIGN KEY ([lock_user_id]) REFERENCES [dbo].[Users]([id])
);
GO

CREATE TABLE [dbfs].[BranchCommits] (
	[id] INT NOT NULL IDENTITY(1,1) PRIMARY KEY,
	[branch_id] INT NOT NULL,
	[commit_id] INT NOT NULL,

	CONSTRAINT [FK_BranchCommits__branch_id__Branches] FOREIGN KEY ([branch_id]) REFERENCES [dbfs].[Branches]([id]) ON DELETE CASCADE,
	CONSTRAINT [FK_BranchCommits__commit_id__Commits] FOREIGN KEY ([commit_id]) REFERENCES [dbfs].[Commits]([id]) ON DELETE CASCADE
);
GO

CREATE NONCLUSTERED INDEX [IX_BranchCommits__branch_id]
ON [dbfs].[BranchCommits]([branch_id] ASC);

CREATE NONCLUSTERED INDEX [IX_BranchCommits__commit_id]
ON [dbfs].[BranchCommits]([commit_id] ASC);

-- optimization table
CREATE TABLE [dbfs].[LatestRevisions] (
	[id] INT NOT NULL IDENTITY(1,1) PRIMARY KEY,

	[branch_id] INT NOT NULL,
	[file_id] INT NOT NULL,
	[revision_id] INT NOT NULL,

	CONSTRAINT [FK_LatestRevisions__branch_id__Branches] FOREIGN KEY ([branch_id]) REFERENCES [dbfs].[Branches]([id]) ON DELETE CASCADE,
	CONSTRAINT [FK_LatestRevisions__file_id__Files] FOREIGN KEY ([file_id]) REFERENCES [dbfs].[Files]([id]) ON DELETE CASCADE,
	CONSTRAINT [FK_LatestRevisions__revision_id__FileRevisions] FOREIGN KEY ([revision_id]) REFERENCES [dbfs].[FileRevisions]([id])
);
GO

CREATE NONCLUSTERED INDEX [IX_LatestRevisions__branch_id]
ON [dbfs].[LatestRevisions]([branch_id] ASC);

CREATE NONCLUSTERED INDEX [IX_LatestRevisions__file_id]
ON [dbfs].[LatestRevisions]([file_id] ASC);
GO

-- Special procedures for complex operations

CREATE PROCEDURE [dbfs].[CreateBranch](@name NVARCHAR(256), @origin_commit_id INT, @user_id INT, @new_branch_id INT OUT)
AS
BEGIN
	BEGIN TRANSACTION;

	SET @new_branch_id = 0;

	DECLARE @parent_branch_id INT;
	SET @parent_branch_id = (
		SELECT TOP 1 [branch_id]
		FROM [dbfs].[BranchCommits]
		WHERE [commit_id] = @origin_commit_id
		ORDER BY [branch_id] ASC);

	IF @parent_branch_id IS NULL
		RAISERROR('The argument origin_commit_id is not valid', 18, 1);

	IF @@Error != 0
	BEGIN
		ROLLBACK TRANSACTION;
		RETURN;
	END

	INSERT INTO [dbfs].[Branches]
		([name], [user_id], [origin_commit_id], [latest_commit_id], [date])
	VALUES
		(@name, @user_id, @origin_commit_id, @origin_commit_id, GETUTCDATE());

	IF @@Error != 0
	BEGIN
		ROLLBACK TRANSACTION;
		RETURN;
	END

	-- if we got here then all should be OK unless somebody deleted parent branch by the moment

	DECLARE @branch_id INT;
	SET @branch_id = (SELECT SCOPE_IDENTITY());

	INSERT INTO [dbfs].[BranchCommits]
		([branch_id], [commit_id])
	SELECT
		@branch_id, [commit_id]
	FROM [dbfs].[BranchCommits]
	WHERE [branch_id] = @parent_branch_id;

	IF @@Error != 0
	BEGIN
		ROLLBACK TRANSACTION;
		RETURN;
	END

	INSERT INTO [dbfs].[LatestRevisions]
		([branch_id], [file_id], [revision_id])
	SELECT
		@branch_id,
		[Results].[file_id],
		[Results].[revision_id]
	FROM (
		SELECT
			[Files].[file_id]
		FROM (
			SELECT DISTINCT
				[FileRevisions].[file_id]
			FROM
				[dbfs].[BranchCommits] AS [BranchCommits]

				INNER JOIN [dbfs].[FileRevisions] AS [FileRevisions]
					ON [BranchCommits].[commit_id] = [FileRevisions].[commit_id]

			WHERE
					([BranchCommits].[branch_id] = @parent_branch_id)
				AND ([BranchCommits].[commit_id] <= @origin_commit_id)
		) AS [Files]
	) AS [TargetFiles]
	OUTER APPLY (
		SELECT TOP (1)
			[TargetRevisions].[id] AS [revision_id],
			[TargetRevisions].[file_id]
		FROM (
			SELECT
				[FileRevisions].[id],
				[FileRevisions].[file_id],
				[FileRevisions].[commit_id]
			FROM
				[dbfs].[BranchCommits] AS [BranchCommits]

				INNER JOIN [dbfs].[FileRevisions] AS [FileRevisions]
					ON [BranchCommits].[commit_id] = [FileRevisions].[commit_id]

			WHERE
					([BranchCommits].[branch_id] = @parent_branch_id)
				AND ([BranchCommits].[commit_id] <= @origin_commit_id)
				AND ([TargetFiles].[file_id] = [FileRevisions].[file_id])
		) AS [TargetRevisions]
		ORDER BY [TargetRevisions].[commit_id] DESC
	) AS [Results];

	IF @@Error != 0
	BEGIN
		ROLLBACK TRANSACTION;
		RETURN;
	END

	COMMIT TRANSACTION;
	SET @new_branch_id = @branch_id;
END
GO

CREATE PROCEDURE [dbfs].[LockBranch](@branch_id INT, @user_id INT, @timeout INT)
AS
BEGIN
	BEGIN TRANSACTION;

	DECLARE @lock_release_date DATETIME;
	SET @lock_release_date = (SELECT [lock_release_date] FROM [dbfs].[Branches] WITH (HOLDLOCK, UPDLOCK) WHERE [id] = @branch_id);

	IF @lock_release_date IS NOT NULL AND DATEDIFF(second, GETUTCDATE(), @lock_release_date) > 0
	BEGIN
		ROLLBACK TRANSACTION;
		RETURN 0;
	END

	UPDATE [dbfs].[Branches]
	SET [lock_user_id] = @user_id,
		[lock_release_date] = DATEADD(second, @timeout, GETUTCDATE())
	WHERE [id] = @branch_id;

	IF @@Error != 0
	BEGIN
		ROLLBACK TRANSACTION;
		RETURN 0;
	END

	COMMIT TRANSACTION;
	RETURN 1;
END
GO

CREATE PROCEDURE [dbfs].[RemoveBranch](@branch_id INT)
AS
BEGIN
	BEGIN TRANSACTION;

	DELETE FROM [dbfs].[Branches]
	WHERE [id] = @branch_id;

	IF @@Error != 0
	BEGIN
		ROLLBACK TRANSACTION;
		RETURN;
	END

	DELETE FROM [dbfs].[Commits]
	WHERE [id] NOT IN
		(SELECT [commit_id]
		FROM [dbfs].[BranchCommits]);

	IF @@Error != 0
	BEGIN
		ROLLBACK TRANSACTION;
		RETURN;
	END

	DELETE FROM [dbfs].[Blobs]
	WHERE [id] NOT IN
		(SELECT [blob_id]
		FROM [dbfs].[FileRevisions]
		WHERE [blob_id] IS NOT NULL);

	IF @@Error != 0
	BEGIN
		ROLLBACK TRANSACTION;
		RETURN;
	END

	COMMIT TRANSACTION;
END
GO
